/* 
   sitecopy, remote web site copy manager.
   Copyright (C) 1998, Joe Orton <joe@orton.demon.co.uk> except where otherwise
   indicated.
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This contains my homebrew FTP client implementation.
 * It ain't that good - notably doesn't do any of the telnet protocol
 * on the PI connection which the RFC says it should    - Joe */
 
/* TODO:
 *  - Better connection handling - ensure for every command that the
 * socket is connected, etc, in a similar way the http_request 
 * function does.
 */

#include <sys/types.h>

#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "common.h"
#include "ftp.h"
#include "protocol.h"
#include "socket.h"

extern enum debug_t debug_level;

char *ftp_error;

int pi_socket; /* The socket for the PI (control) connection */
int dtp_socket; /* The socket for the DTP (data) connection */

bool ftp_use_passive = true; /* whether we use PASV mode or not */

/* The address of the server for the DTP connection - 
 * this goes here because it's got from ftp_read() rather than in a decent
 * manner */
unsigned short dtp_port;
struct in_addr dtp_addr;

struct sockaddr_in dtp_host;

struct proto_driver ftp_driver = {
    ftp_connect,
    ftp_login,
    ftp_close,
    ftp_move,
    ftp_put,
    ftp_delete,
    ftp_chmod,
    ftp_mkdir,
    ftp_rmdir,
    &ftp_error,
    "ftp",
    "FTP"
};

int ftp_close( ) {
    return ftp_exec( "QUIT" );
}

/* Changes to given directory
 * Gives appropriate FTP state response */
int ftp_chdir( const char *dir ) {
    char command[BUFSIZ];
    snprintf( command, BUFSIZ, "CWD %s", dir );
    return ftp_exec( command );
}
 
/* Creates the given directory
 * FTP state response */
int ftp_mkdir( const char *dir ) {
    char command[BUFSIZ];
    snprintf( command, BUFSIZ, "MKD %s", dir );
    return ftp_exec( command );
}
 
/* Renames or moves a file */
int ftp_move( const char *from, const char *to ) {
    char command[BUFSIZ];
    snprintf( command, BUFSIZ, "RNFR %s", from );
    if( ftp_exec( command ) == FTP_FILEMORE ) {
	snprintf( command, BUFSIZ, "RNTO %s", to );
	return ftp_exec( command );
    } else {
	return FTP_ERROR;
    }
}

int ftp_delete( const char *filename ) {
    char command[BUFSIZ];
    snprintf( command, BUFSIZ, "DELE %s", filename );
    return ftp_exec( command );
}

int ftp_rmdir( const char *filename ) {
    char command[BUFSIZ];
    snprintf( command, BUFSIZ, "RMD %s", filename );
    return ftp_exec( command );
}

int ftp_namelist( const char *dir, char **list ) {
    int ret, len, wrlen;
    char command[BUFSIZ], buffer[BUFSIZ];
    snprintf( command, BUFSIZ, "NLST %s", dir );
    ret = ftp_data_open( command );
    switch( ret ) {
    case FTP_READY:
	/* Read the listing */
	do {
	    len = read( dtp_socket, buffer, BUFSIZ );
	    if( len < 0 ) perror("read");
	    if( len > 0 ) wrlen = write( fileno(stdout), buffer, len );
	} while( len > 0 );
	fflush( stdout );
	return ftp_data_close( );
    case FTP_NOPASSIVE:
	/* No passive allowed */
	printf( "Server does not support passive mode - cannot connect.\n" );
	break;
    default:
	printf( "Could not retrieve directory listing: %d\n", ret );
	close( dtp_socket );
	break;
    }
    return ret;
}

/* Changes to given directory, creating if necessary and possible */
int ftp_godir( const char *dir ) {
    int ret;
    ret = ftp_chdir( dir ) != FTP_OK;
    if( ret != FTP_OK ) {
	/* Attempt to create... */
	ret = ftp_mkdir( dir );
	if( ret == FTP_OK )
	    /* Attempt to chdir... */
	    ret = ftp_chdir( dir );
    }
    return ret;
}

/* FTP non-PASV mode open */

int ftp_active_open(const char *command) {
  char *a, *p;
  struct sockaddr_in myctladdr;
  char cmd[BUFSIZ];
  int len, ret;
  int sock_connected;
  struct sockaddr_in from;
  size_t fromlen = sizeof(from);

  /* Get the address of the near end of the control connection.
   */
  if (getsockname(pi_socket, (struct sockaddr *)&myctladdr, &len) < 0) {
    perror("ftp: getsockname");
    return 0;
  }

  /* Our local address is the same for both data and control connections.
   */
  dtp_host = myctladdr;
  dtp_host.sin_port = htons(0);	/* pick one for us. */
  
  /* Create a local socket.
   */  
  if ( (dtp_socket = socket( AF_INET, SOCK_STREAM, 0 )) < 0 ) {
    perror("socket");
    return 0;
  }

  /* Bind it to a port.
   */
  if (0 != bind(dtp_socket, (struct sockaddr *)&dtp_host, sizeof(dtp_host))) {
    perror("bind");
    return 0;
  }

  /* Get the name of the bound (data) socket.
   */
  if (getsockname(dtp_socket, (struct sockaddr *)&dtp_host, &len) < 0) {
    perror("ftp: getsockname");
  }

  if (listen(dtp_socket, 1) < 0) {
    perror("listen");
    return 0;
  }
  


  /* Let the remote end know the address of our port.
   */
#define	UC(b)	(((int)b)&0xff)
  a = (char *)&dtp_host.sin_addr;
  p = (char *)&dtp_host.sin_port;
  
  snprintf(cmd, BUFSIZ, "PORT %d,%d,%d,%d,%d,%d",
		      UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]),
		      UC(p[0]), UC(p[1]));
  ret = ftp_exec(cmd);

  /* Send the command.  This will make the remote end
   * initiate the connection.
   */
  ret = ftp_exec(command);

  /* Now wait for a connection from the remote end.
   */
  sock_connected = accept(dtp_socket, (struct sockaddr *)&from, &fromlen);
  if (sock_connected < 0) {
    perror("accept");
  }
  close(dtp_socket);
  dtp_socket = sock_connected;

  return ret;
}



/* Establishes the control connection, to hostname on portnum */
int ftp_connect( const char *hostname, const int portnum ) {
    struct in_addr addr;
    if( host_lookup( hostname, &addr ) )
	return -1;
    pi_socket = socket_connect( addr, portnum );
    /* Skip everything the hello message */
    return ftp_read();
}

int ftp_chmod( const char *filename, const mode_t mode ) {
    char command[BUFSIZ];
    snprintf( command, BUFSIZ, "SITE CHMOD %3o %s", mode & 0777, filename );
    return ftp_exec( command );
}

int ftp_data_open( const char *command ) {
    int ret;
    if( ftp_use_passive ) {
	ret = ftp_exec( "PASV" );
	if( ret == FTP_PASSIVE ) {
	    if( ftp_connect_pasv( ) ) {
		return ftp_exec( command );
	    } else {
		return FTP_ERROR;
	    }
	} else {
	    return FTP_NOPASSIVE;
	}
    } else {
	/* we are not using passive mode. */
	return ftp_active_open(command);
    }

}

/* Closes the data connection */
int ftp_data_close( ) {
    close( dtp_socket );
    /* sent okay? */
    return ftp_read();
}

/* upload the given file */
int ftp_put( const char *localfile, const char *remotefile ) {
    int ret;
    char command[BUFSIZ];
    strcpy( command, "TYPE I" );
    ftp_exec( command );
    snprintf( command, BUFSIZ, "STOR %s", remotefile );
    ret = ftp_data_open( command );
    switch( ret ) {
    case FTP_READY:
	/* send the file */
	send_file( dtp_socket, localfile );
	return ftp_data_close( );
	break;
    case FTP_NOPASSIVE:
	/* No passive allowed */
	printf( "Server does not support passive mode - cannot connect.\n" );
	break;
    default:
	printf( "Could not send file.\n" );
	close( dtp_socket );
	break;
    }
    return ret;
}

/* Connect to the given host 
 * Returns non-zero value if successfull */
int ftp_connect_pasv() {
    dtp_socket = socket_connect( dtp_addr, dtp_port );
    return 1;
}


/* Do a login */
int ftp_login( const char *username, const char *password ) {
    char cmd[BUFSIZ];
    int ret;
    debug( debug_some,  "Logging in as %s...\n", username);
    snprintf( cmd, BUFSIZ, "USER %s", username );
    ret = ftp_exec( cmd );
    if( ret == FTP_NEEDPASSWORD ) {
	debug( debug_some,  "Supplying password...\n" );
	snprintf( cmd, BUFSIZ, "PASS %s", password );
	ret = ftp_exec( cmd );
    } else {
	debug( debug_some, "Password not required.\n");
    }
    return ret;
}

int ftp_exec( const char *command ) {
    if( debug_level > debug_none)
	fprintf( stderr, ">%s\n", command );
    send_line( pi_socket, (char *) command );
    return ftp_read();
}

int ftp_read() {
    int multiline, len, reply_code;
    char buffer[BUFSIZ];

    multiline = 0;

    for(;;) {
	if( (len = read_line(pi_socket, buffer, BUFSIZ )) < 0 ) {
	    fprintf( stderr, "Could not read from socket." );
	    break;
	}
	
	debug( debug_some, "<%s", buffer );
	if(len<5) /* this is a multi-liner, ignore it and carry on */
	    continue; 
	/* parse the reply code from the line */
	reply_code = get_reply_code(buffer);
	
	/* If we have a VALID reply code and we are currently
	 * in a multi line response then this is the end of the
	 * multi line response */
	if(multiline && reply_code)
	    multiline=0;
	
	/* Now, if we aren't in a multi line response... */
	if(!multiline) { 
	    if(buffer[3]=='-') {
		/* A dash in char four denotes beginning of multi
		 * line response  'xxx-' */
		multiline=1;
	    } else {
		/* Looks like we've got a real response line */
		return ftp_response( buffer, reply_code );
	    }
	}
    }
    return FTP_ERROR;
}

int ftp_response( const char *response, const int code ) {
    char *newline;
    if( ftp_error!=NULL) free(ftp_error);
    ftp_error = strdup( response );
    /* Chop the newline */
    newline = strchr( ftp_error, '\n' );
    if( newline ) *newline = '\0';	
    switch( code ) {
    case 200: /* misc OK codes */
    case 220:
    case 230:
    case 250: /* completed file action */
    case 257: /* mkdir success */
    case 226: /* received file okay */
	return FTP_OK;
    case 150: /* ready for data */
	return FTP_READY;
    case 550: /* couldn't complete file action */
	return FTP_FILEBAD;
    case 331: /* got username, want password */
	return FTP_NEEDPASSWORD;
    case 350: /* file action pending further info - RNFR */
	return FTP_FILEMORE;
    case 221:
	return FTP_CLOSED;
    case 421: /* service denied */
	return FTP_DENIED;
    case 227: /* Passive mode */
	if( ftp_read_pasv( response ) )
	    return FTP_PASSIVE;
	else
	    return FTP_ERROR;
    case 553: /* couldn't create directory */
	return FTP_ERROR;
    default:
	return FTP_ERROR;
    }
}

int ftp_read_pasv( const char *response ) {
    int h1, h2, h3, h4, p1, p2;
    char *start;
    start = strchr( response, '(' );
    /* get the host + port */
    if( sscanf( ++start, "%d,%d,%d,%d,%d,%d", &h1, &h2, &h3, &h4, &p1, &p2 ) < 6 )
	/* didn't match, give up */
	return 0;
    /* Record this for future reference */
    dtp_port = (p1<<8) | p2;
    dtp_addr.s_addr = htonl( (h1<<24) | (h2<<16) | (h3<<8) | h4 );
    return 1;
}

int get_reply_code( const char *buffer) {
    if(strlen(buffer) > 3)
	if( isdigit(buffer[0]) && isdigit(buffer[1]) && isdigit(buffer[2]) )
	    /* looks good */
	    return atoi(buffer);
    return 0;
}
